// -*- c++ -*-
/*
 *  MICO --- a CORBA 2.0 implementation
 *  Copyright (C) 1997 Kay Roemer & Arno Puder
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#ifndef __mico_dii_h__
#define __mico_dii_h__


enum MemoryFlags {
    IN_COPY_VALUE = 0x0001,
    OUT_LIST_MEMORY = 0x0002,
    DEPENDENT_LIST = 0x0004,
};

// environment
class Environment : public ServerlessObject {
    Exception *except;
    Principal_ptr princ;
public:
    void exception (Exception *);
    Exception *exception (); // XXX const;
    void clear ();

    // begin-mico-extension
    Environment (Principal_ptr = NULL);
    ~Environment ();

    Principal_ptr principal ();
    // end-mico-extension

    static Environment_ptr _duplicate (Environment_ptr);
    static Environment_ptr _nil ();
};
MICO_EXPORT_DECL void release (Environment_ptr);
MICO_EXPORT_DECL Boolean is_nil (Environment_ptr);

typedef ObjVar<Environment> Environment_var;


// named value
class NamedValue : public ServerlessObject {
    string n;
    Any v;
    Flags f;
public:
    const char *name () const;
    Any *value (); // XXX const doesn't work here
    Flags flags () const;

    // begin-mico-extension
    NamedValue (const char *name, const Any &, Flags f = 0);
    NamedValue (const char *name, Flags f = 0);
    NamedValue (); // vector<NamedValue> needs this
    NamedValue &operator= (const NamedValue &);
    ~NamedValue ();
    // end-mico-extension

    static NamedValue_ptr _duplicate (NamedValue_ptr);
    static NamedValue_ptr _nil ();
};
MICO_EXPORT_DECL void release (NamedValue_ptr);
MICO_EXPORT_DECL Boolean is_nil (NamedValue_ptr);

typedef ObjVar<NamedValue> NamedValue_var;


// nv list
class NVList : public ServerlessObject {
    ULong cnt;
    vector<NamedValue_var> vec;
public:
    ULong count () const;
    NamedValue_ptr add (Flags);
    NamedValue_ptr add_item (const char *, Flags);
    NamedValue_ptr add_value (const char *, const Any &, Flags);
    NamedValue_ptr add_item_consume (char *, Flags);
    NamedValue_ptr add_value_consume (char *, Any *, Flags);
    NamedValue_ptr item (ULong idx); // XXX raises Bounds
    Status remove (ULong); // XXX raises Bounds

    // begin-mico-extension
    NVList (ULong count = 0);
    ~NVList ();

    void copy (NVList_ptr, Flags flags);
    // end-mico-extension

    static NVList_ptr _duplicate (NVList_ptr);
    static NVList_ptr _nil ();
};
MICO_EXPORT_DECL void release (NVList_ptr);
MICO_EXPORT_DECL Boolean is_nil (NVList_ptr);

typedef ObjVar<NVList> NVList_var;


// exception list
class ExceptionList : public ServerlessObject {
    vector<TypeCode_var> vec;

public:
    ULong count () const;
    void add (TypeCode_ptr tc);
    void add_consume (TypeCode_ptr tc);
    TypeCode_ptr item (ULong idx);
    Status remove (ULong idx);

    // begin-mico-extension
    ExceptionList ();
    ~ExceptionList ();
    // end-mico-extension

    static ExceptionList_ptr _duplicate (ExceptionList_ptr);
    static ExceptionList_ptr _nil ();
};
MICO_EXPORT_DECL void release (ExceptionList_ptr);
MICO_EXPORT_DECL Boolean is_nil (ExceptionList_ptr);

typedef ObjVar<ExceptionList> ExceptionList_var;


// context list
class ContextList : public ServerlessObject {
    vector<string> vec;

public:
    ULong count () const;
    void add (const char *ctxt);
    void add_consume (char *ctxt);
    const char *item (ULong idx);
    Status remove (ULong idx);

    // begin-mico-extension
    ContextList ();
    ~ContextList ();
    // end-mico-extension

    static ContextList_ptr _duplicate (ContextList_ptr);
    static ContextList_ptr _nil ();
};
MICO_EXPORT_DECL void release (ContextList_ptr);
MICO_EXPORT_DECL Boolean is_nil (ContextList_ptr);

typedef ObjVar<ContextList> ContextList_var;


enum ArgModeFlags {
    ARG_IN = 0x0001,
    ARG_OUT = 0x0002,
    ARG_INOUT = 0x0004,
};

enum InvocationFlags {
    INV_NO_RESPONSE = 0x0001,
    INV_TERM_ON_ERROR = 0x0002,
    RESP_NO_WAIT = 0x0004,
};

struct RequestCallback {
    enum Event { RequestDone };
    virtual void callback (Request_ptr, Event) = 0;
    virtual ~RequestCallback ();
};

// request
class Request : public ServerlessObject, public ORBCallback {
    typedef ULong MsgId; // XXX should be ORB::MsgId

    Object_ptr object;
    Context_ptr context;
    string opname;
    NVList_ptr args;
    NamedValue_ptr res;
    Flags flags;
    ExceptionList_ptr elist;
    ContextList_ptr clist;
    Environment_ptr environ;
    MsgId msgid;
    Boolean invoke_pending;
    ORBRequest *orbreq;
    RequestCallback *cb;

    Request (const Request &) { assert (0); }
    Request &operator= (const Request &) { assert (0); return *_nil(); }

    void decode_user_exception ();

    virtual void callback (ORB_ptr, ULong, ORBCallback::Event);
public:
    Object_ptr target (); // XXX const;
    const char *operation () const;
    NVList_ptr arguments ();
    NamedValue_ptr result ();
    Environment_ptr env ();
    ExceptionList_ptr exceptions ();
    ContextList_ptr contexts ();
    
    void ctx (Context_ptr);
    Context_ptr ctx (); // XXX const;

    Status invoke ();
    Status send_oneway ();
    Status send_deferred (RequestCallback *cb = 0);
    Status get_response (Boolean block = TRUE);
    Boolean poll_response ();

    Any &add_in_arg ();
    Any &add_in_arg (const char *name);
    Any &add_inout_arg ();
    Any &add_inout_arg (const char *name);
    Any &add_out_arg ();
    Any &add_out_arg (const char *name);
    void set_return_type (TypeCode_ptr);
    Any &return_value ();

    // begin-mico-extension
    Request (Object_ptr, Context_ptr, const char *op,
	     NVList_ptr args, NamedValue_ptr result,
	     Flags f);
    Request (Object_ptr, Context_ptr, const char *op,
	     NVList_ptr args, NamedValue_ptr result,
	     ExceptionList_ptr, ContextList_ptr,
	     Flags f);
    Request (Object_ptr, const char *op);
    Request (); // vector<Request> needs this
    ~Request ();
    // end-mico-extension

    static Request_ptr _duplicate (Request_ptr);
    static Request_ptr _nil ();
};
MICO_EXPORT_DECL void release (Request_ptr);
MICO_EXPORT_DECL Boolean is_nil (Request_ptr);

typedef ObjVar<Request> Request_var;


enum ContextFlags {
    CTX_DELETE_DESCENDENTS = 0x0001,
    CTX_RESTRICT_SCOPE = 0x0002,
};

typedef ObjVar<Context> Context_var;

// context
class Context : public ServerlessObject {
    Context_ptr myparent;
    vector<Context_var> childs;
    NVList_var properties;
    string name;

    Boolean match (const char *value, const char *pattern);
public:
    const char *context_name () const;
    Context_ptr parent (); // XXX const;

    Status create_child (const char *, Context_ptr &);

    Status set_one_value (const char *, const Any &);
    Status set_values (NVList_ptr);
    Status delete_values (const char *);
    Status get_values (const char *, Flags, const char *, NVList_ptr &);
    Status get_values (const char *, Flags, ContextList_ptr, NVList_ptr &);

    // begin-mico-extension
    Context (DataDecoder &);
    Context (const char *name, Context_ptr parent = 0);
    ~Context ();

    void encode (DataEncoder &, ContextList_ptr = 0) const;
    Boolean decode (DataDecoder &);
    // end-mico-extension

    static Context_ptr _duplicate (Context_ptr);
    static Context_ptr _nil ();
};
MICO_EXPORT_DECL void release (Context_ptr);
MICO_EXPORT_DECL Boolean is_nil (Context_ptr);


// principal
class Principal : public ServerlessObject {
    vector<Octet> rep;

public:
    // begin-mico-extension
    Principal (DataDecoder &);
    Principal ();
    ~Principal ();

    void encode (DataEncoder &) const;
    Boolean decode (DataDecoder &);

    Boolean operator== (const Principal &) const;
    // end-mico-extension

    static Principal_ptr _duplicate (Principal_ptr);
    static Principal_ptr _nil ();
};
MICO_EXPORT_DECL void release (Principal_ptr);
MICO_EXPORT_DECL Boolean is_nil (Principal_ptr);

typedef ObjVar<Principal> Principal_var;

#endif // __mico_dii_h__
