/*
 *  MICO --- a CORBA 2.0 implementation
 *  Copyright (C) 1997 Kay Roemer & Arno Puder
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */


#ifndef __IR_IMPL_H__
#define __IR_IMPL_H__


#define MICO_IR_EXC_DUPLICATE_NAME   100
#define MICO_IR_EXC_DUPLICATE_REPOID 101
#define MICO_IR_EXC_DUPLICATE_LABEL  102
#define MICO_IR_EXC_AMBIGUOUS_NAME   103
#define MICO_IR_EXC_OVERLOADING      104
#define MICO_IR_EXC_BAD_RECURSION    105
#define MICO_IR_EXC_BAD_ONEWAY       106
#define MICO_IR_EXC_DUPLICATE_MEMBER 107



class IRObject_impl : virtual public CORBA::IRObject_skel
{
protected:
  CORBA::DefinitionKind _dk;

  void raise_exception( CORBA::ULong err );
  
public:
  IRObject_impl();
  virtual CORBA::DefinitionKind def_kind();
  virtual void destroy();
};


class Contained_impl :
  virtual public IRObject_impl,
  virtual public CORBA::Contained_skel
{
protected:
  CORBA::RepositoryId_var _id;
  CORBA::Identifier_var   _name;
  CORBA::VersionSpec_var  _version;
  CORBA::Container_var    _defined_in;

  CORBA::RepositoryId container_id ();
public:
  Contained_impl( CORBA::Container_ptr defined_in );
  virtual CORBA::RepositoryId id();
  virtual void id( const CORBA::RepositoryId value );
  virtual CORBA::Identifier name();
  virtual void name( const CORBA::Identifier value );
  virtual CORBA::VersionSpec version();
  virtual void version( const CORBA::VersionSpec value );
  virtual CORBA::Container_ptr defined_in();
  virtual CORBA::ScopedName absolute_name();
  virtual CORBA::ScopedName get_absolute_name(const char *separator);
  virtual CORBA::Repository_ptr containing_repository();
  virtual CORBA::Contained::Description* describe();
  virtual void move( CORBA::Container_ptr new_container,
		     const CORBA::Identifier new_name,
		     const CORBA::VersionSpec new_version );
};


class Container_impl : 
  virtual public IRObject_impl,
  virtual public CORBA::Container_skel
{
protected:
  CORBA::ContainedSeq _contents;

  void check_for_duplicates( const CORBA::Identifier name,
			     const CORBA::RepositoryId id );
  bool is_scoped_name( string& name );
  bool is_relative_scoped_name( string& name );
  string first_of_scoped_name( string& name );
  void strip_first_scope( string& name );
  CORBA::Contained_ptr locate_id( const CORBA::RepositoryId id );
  CORBA::Contained_ptr locate_name( const CORBA::Identifier name,
				    CORBA::Boolean include_enclosing_scopes,
				    CORBA::Boolean is_first_level );
  
public:
  Container_impl();
  virtual CORBA::Contained_ptr lookup( CORBA::ScopedName search_name );
  virtual CORBA::ContainedSeq* contents( CORBA::DefinitionKind limit_type,
					 CORBA::Boolean exclude_inherited );
  virtual CORBA::ContainedSeq* lookup_name( CORBA::Identifier search_name,
					    CORBA::Long levels_to_search,
					    CORBA::DefinitionKind limit_type,
					    CORBA::Boolean exclude_inherited );
  virtual CORBA::Container::DescriptionSeq* 
                     describe_contents( CORBA::DefinitionKind limit_type,
					CORBA::Boolean exclude_inherited,
					CORBA::Long max_returned_objs );
  virtual CORBA::ModuleDef_ptr create_module( const CORBA::RepositoryId id,
					      const CORBA::Identifier name,
					      const CORBA::VersionSpec version );
  virtual CORBA::ConstantDef_ptr create_constant( const CORBA::RepositoryId id,
						  const CORBA::Identifier name,
						  const CORBA::VersionSpec version,
						  CORBA::IDLType_ptr type,
						  const CORBA::Any& value );
  virtual CORBA::ExceptionDef_ptr create_exception( const CORBA::RepositoryId id,
						    const CORBA::Identifier name,
						    const CORBA::VersionSpec version,
						    const CORBA::StructMemberSeq& members );
  virtual CORBA::StructDef_ptr create_struct( const CORBA::RepositoryId id,
					      const CORBA::Identifier name,
					      const CORBA::VersionSpec version,
					      const CORBA::StructMemberSeq& members );
  virtual CORBA::UnionDef_ptr create_union( const CORBA::RepositoryId id,
					    const CORBA::Identifier name,
					    const CORBA::VersionSpec version,
					    CORBA::IDLType_ptr discriminator_type,
					    const CORBA::UnionMemberSeq& members );
  virtual CORBA::EnumDef_ptr create_enum( const CORBA::RepositoryId id,
					  const CORBA::Identifier name,
					  const CORBA::VersionSpec version,
					  const CORBA::EnumMemberSeq& members );
  virtual CORBA::AliasDef_ptr create_alias( const CORBA::RepositoryId id,
					    const CORBA::Identifier name,
					    const CORBA::VersionSpec version,
					    CORBA::IDLType_ptr original_type );
  virtual CORBA::InterfaceDef_ptr create_interface( const CORBA::RepositoryId id,
						    const CORBA::Identifier name,
						    const CORBA::VersionSpec version,
						    const CORBA::InterfaceDefSeq& base_interfaces );
  virtual void remove_contained( CORBA::Contained_ptr con );
  virtual void add_contained( CORBA::Contained_ptr con );
  virtual void insert_struct_or_union( CORBA::Contained_ptr type );
};



class IDLType_impl : 
  virtual public IRObject_impl,
  virtual public CORBA::IDLType_skel
{
protected:
  CORBA::TypeCode_var _type;
  
public:
  IDLType_impl();
  IDLType_impl( CORBA::TypeCode_ptr type );
  virtual CORBA::TypeCode_ptr type();
};


//XXXtypedef SequenceTmpl<CORBA::IDLType_var> IDLTypeSeq;
  
class Repository_impl : 
  virtual public Container_impl,
  virtual public CORBA::Repository_skel
{
protected:
  CORBA::PrimitiveDef_var _pk_void;
  CORBA::PrimitiveDef_var _pk_float;
  CORBA::PrimitiveDef_var _pk_double;
  CORBA::PrimitiveDef_var _pk_long;
  CORBA::PrimitiveDef_var _pk_short;
  CORBA::PrimitiveDef_var _pk_ulong;
  CORBA::PrimitiveDef_var _pk_ushort;
  CORBA::PrimitiveDef_var _pk_char;
  CORBA::PrimitiveDef_var _pk_boolean;
  CORBA::PrimitiveDef_var _pk_octet;
  CORBA::PrimitiveDef_var _pk_string;
  CORBA::PrimitiveDef_var _pk_any;
  CORBA::PrimitiveDef_var _pk_objref;
  CORBA::PrimitiveDef_var _pk_typecode;
  CORBA::PrimitiveDef_var _pk_principal;

  CORBA::IDLTypeSeq _anonymous_types;
  
  CORBA::IDLType_ptr add_anonymous_type( CORBA::IDLType_ptr type );

public:
  Repository_impl();
  virtual CORBA::Contained_ptr lookup_id( const CORBA::RepositoryId search_id );
  virtual CORBA::PrimitiveDef_ptr get_primitive( CORBA::PrimitiveKind kind );
  virtual CORBA::StringDef_ptr create_string( CORBA::ULong bound );
  virtual CORBA::SequenceDef_ptr create_sequence( CORBA::ULong bound,
						   CORBA::IDLType_ptr element_type );
  virtual CORBA::ArrayDef_ptr create_array( CORBA::ULong length,
					     CORBA::IDLType_ptr element_type );
  virtual CORBA::IDLTypeSeq *get_anonymous_types();
};


class ModuleDef_impl : 
  virtual public Container_impl,
  virtual public Contained_impl,
  virtual public CORBA::ModuleDef_skel
{
public:
  ModuleDef_impl( CORBA::Container_ptr defined_in,
		  const CORBA::RepositoryId id,
		  const CORBA::Identifier name,
		  const CORBA::VersionSpec version );
};


class ConstantDef_impl : 
  virtual public Contained_impl,
  virtual public CORBA::ConstantDef_skel
{
protected:
  CORBA::IDLType_var _type_def;
  CORBA::Any         _value;
public:
  ConstantDef_impl( CORBA::Container_ptr defined_in,
		    const CORBA::RepositoryId id,
		    const CORBA::Identifier name,
		    const CORBA::VersionSpec version,
		    CORBA::IDLType_ptr type,
		    const CORBA::Any &value );
  virtual CORBA::TypeCode_ptr type();
  virtual CORBA::IDLType_ptr type_def();
  virtual void type_def( CORBA::IDLType_ptr value );
  virtual CORBA::Any* value();
  virtual void value( const CORBA::Any& value );
};



class TypedefDef_impl : 
  virtual public Contained_impl,
  virtual public IDLType_impl,
  virtual public CORBA::TypedefDef_skel
{
public:
  TypedefDef_impl( CORBA::Container_ptr defined_in );
};



class StructDef_impl : 
  virtual public TypedefDef_impl,
  virtual public CORBA::StructDef_skel
{
protected:
  CORBA::StructMemberSeq _tmp_members;
  CORBA::StructMemberSeq _members;

private:
  void computeIDLTypes();
  void check_for_bad_recursion( CORBA::UnionMemberSeq& mem );
  void check_for_bad_recursion( CORBA::StructMemberSeq& mem );
  
public:
  StructDef_impl( CORBA::Container_ptr defined_in,
		  const CORBA::RepositoryId id,
		  const CORBA::Identifier name,
		  const CORBA::VersionSpec version,
		  const CORBA::StructMemberSeq& members );
  virtual CORBA::StructMemberSeq* members();
  virtual void members( const CORBA::StructMemberSeq& value );
};



class UnionDef_impl : 
  virtual public TypedefDef_impl,
  virtual public CORBA::UnionDef_skel
{
protected:
  CORBA::IDLType_var    _discr;
  CORBA::UnionMemberSeq _tmp_members;
  CORBA::UnionMemberSeq _members;

private:
  void computeIDLTypes();
  void check_for_bad_recursion( CORBA::UnionMemberSeq& mem );
  void check_for_bad_recursion( CORBA::StructMemberSeq& mem );

public:
  UnionDef_impl( CORBA::Container_ptr defined_in,
		 const CORBA::RepositoryId id,
		 const CORBA::Identifier name,
		 const CORBA::VersionSpec version,
		 CORBA::IDLType_ptr discr,
		 const CORBA::UnionMemberSeq& members );
  virtual CORBA::TypeCode_ptr discriminator_type();
  virtual CORBA::IDLType_ptr discriminator_type_def();
  virtual void discriminator_type_def( CORBA::IDLType_ptr value );
  virtual CORBA::UnionMemberSeq* members();
  virtual void members( const CORBA::UnionMemberSeq& value );
};



class EnumDef_impl : 
  virtual public TypedefDef_impl,
  virtual public CORBA::EnumDef_skel
{
protected:
  CORBA::EnumMemberSeq _members;
public:
  EnumDef_impl( CORBA::Container_ptr defined_in,
		const CORBA::RepositoryId id,
		const CORBA::Identifier name,
		const CORBA::VersionSpec version );
  virtual CORBA::EnumMemberSeq* members();
  virtual void members( const CORBA::EnumMemberSeq& value );
};



class AliasDef_impl : 
  virtual public TypedefDef_impl,
  virtual public CORBA::AliasDef_skel
{
protected:
  CORBA::IDLType_var _original;
public:
  AliasDef_impl( CORBA::Container_ptr defined_in,
		 const CORBA::RepositoryId id,
		 const CORBA::Identifier name,
		 const CORBA::VersionSpec version );
  virtual CORBA::IDLType_ptr original_type_def();
  virtual void original_type_def( CORBA::IDLType_ptr value );
};


class PrimitiveDef_impl : 
  virtual public IDLType_impl,
  virtual public CORBA::PrimitiveDef_skel
{
protected:
  CORBA::PrimitiveKind _kind;
public:
  PrimitiveDef_impl( CORBA::PrimitiveKind kind );
  virtual CORBA::PrimitiveKind kind();
};



class StringDef_impl : 
  virtual public IDLType_impl,
  virtual public CORBA::StringDef_skel
{
protected:
  CORBA::ULong _bound;
public:
  StringDef_impl();
  virtual CORBA::ULong bound();
  virtual void bound( CORBA::ULong value );
};



class SequenceDef_impl : 
  virtual public IDLType_impl,
  virtual public CORBA::SequenceDef_skel
{
protected:
  CORBA::ULong       _bound;
  CORBA::IDLType_var _element_type_def;

  // Offset for recursive sequence
  CORBA::ULong _offset;

  CORBA::TypeCode_ptr lookup_recursive_tc( CORBA::TypeCode_ptr tc,
					   CORBA::ULong depth );

  void compute_tc();
  void compute_recursive_tc();
  
public:
  SequenceDef_impl();
  virtual CORBA::TypeCode_ptr type();
  virtual CORBA::ULong bound();
  virtual void bound( CORBA::ULong value );
  virtual CORBA::TypeCode_ptr element_type();
  virtual CORBA::IDLType_ptr element_type_def();
  virtual void element_type_def( CORBA::IDLType_ptr value );
};



class ArrayDef_impl : 
  virtual public IDLType_impl,
  virtual public CORBA::ArrayDef_skel
{
protected:
  CORBA::ULong       _length;
  CORBA::IDLType_var _element_type_def;

  void compute_tc();
  
public:
  ArrayDef_impl();
  virtual CORBA::ULong length();
  virtual void length( CORBA::ULong value );
  virtual CORBA::TypeCode_ptr element_type();
  virtual CORBA::IDLType_ptr element_type_def();
  virtual void element_type_def( CORBA::IDLType_ptr value );
};



class ExceptionDef_impl : 
  virtual public Contained_impl,
  virtual public CORBA::ExceptionDef_skel
{
protected:
  CORBA::StructMemberSeq _members;
  CORBA::TypeCode_var    _type;
private:
  void computeIDLTypes();
public:
  ExceptionDef_impl( CORBA::Container_ptr defined_in,
		     const CORBA::RepositoryId id,
		     const CORBA::Identifier name,
		     const CORBA::VersionSpec version,
		     const CORBA::StructMemberSeq& members );
  virtual CORBA::TypeCode_ptr type();
  virtual CORBA::StructMemberSeq* members();
  virtual void members( const CORBA::StructMemberSeq& value );
};



class AttributeDef_impl : 
  virtual public Contained_impl,
  virtual public CORBA::AttributeDef_skel
{
protected:
  CORBA::IDLType_var   _type_def;
  CORBA::AttributeMode _mode;
  
public:
  AttributeDef_impl( CORBA::Container_ptr defined_in,
		     const CORBA::RepositoryId id,
		     const CORBA::Identifier name,
		     const CORBA::VersionSpec version,
		     CORBA::IDLType_ptr type_def,
		     CORBA::AttributeMode mode );
  virtual CORBA::TypeCode_ptr type();
  virtual CORBA::IDLType_ptr type_def();
  virtual void type_def( CORBA::IDLType_ptr value );
  virtual CORBA::AttributeMode mode();
  virtual void mode( CORBA::AttributeMode value );
};


class OperationDef_impl : 
  virtual public Contained_impl,
  virtual public CORBA::OperationDef_skel
{
protected:
  CORBA::IDLType_var       _result_def;
  CORBA::ParDescriptionSeq _params;
  CORBA::OperationMode     _mode;
  CORBA::ExceptionDefSeq   _exceptions;
  CORBA::ContextIdSeq      _contexts;
  
public:
  OperationDef_impl( CORBA::Container_ptr defined_in,
		     const CORBA::RepositoryId id,
		     const CORBA::Identifier name,
		     const CORBA::VersionSpec version,
		     CORBA::IDLType_ptr result,
		     CORBA::OperationMode mode,
		     const CORBA::ParDescriptionSeq& params,
		     const CORBA::ExceptionDefSeq& exs,
		     const CORBA::ContextIdSeq& ctx );
  virtual CORBA::TypeCode_ptr result();
  virtual CORBA::IDLType_ptr result_def();
  virtual void result_def( CORBA::IDLType_ptr value );
  virtual CORBA::ParDescriptionSeq* params();
  virtual void params( const CORBA::ParDescriptionSeq& value );
  virtual CORBA::OperationMode mode();
  virtual void mode( CORBA::OperationMode value );
  virtual CORBA::ContextIdSeq* contexts();
  virtual void contexts( const CORBA::ContextIdSeq& value );
  virtual CORBA::ExceptionDefSeq* exceptions();
  virtual void exceptions( const CORBA::ExceptionDefSeq& value );
};



class InterfaceDef_impl : 
  virtual public Container_impl,
  virtual public Contained_impl,
  virtual public IDLType_impl,
  virtual public CORBA::InterfaceDef_skel
{
protected:
  CORBA::InterfaceDefSeq _base_interfaces;
  
  void check_for_overloaded_ops( const CORBA::Identifier name,
				 const CORBA::RepositoryId id );
public:
  InterfaceDef_impl();
  InterfaceDef_impl( CORBA::Container_ptr defined_in,
		     const CORBA::RepositoryId id,
		     const CORBA::Identifier name,
		     const CORBA::VersionSpec version );
  virtual CORBA::InterfaceDefSeq* base_interfaces();
  virtual void base_interfaces( const CORBA::InterfaceDefSeq& value );
  virtual CORBA::Boolean is_a( const CORBA::RepositoryId interface_id );
  virtual CORBA::InterfaceDef::FullInterfaceDescription* describe_interface();
  virtual CORBA::AttributeDef_ptr create_attribute( const CORBA::RepositoryId id,
						    const CORBA::Identifier name,
						    const CORBA::VersionSpec version,
						    CORBA::IDLType_ptr type,
						    CORBA::AttributeMode mode );
  virtual CORBA::OperationDef_ptr create_operation( const CORBA::RepositoryId id,
						    const CORBA::Identifier name,
						    const CORBA::VersionSpec version,
						    CORBA::IDLType_ptr result,
						    CORBA::OperationMode mode,
						    const CORBA::ParDescriptionSeq& params,
						    const CORBA::ExceptionDefSeq& exceptions,
						    const CORBA::ContextIdSeq& contexts );
};


#endif
