// -*- c++ -*-
/*
 *  MICO --- a CORBA 2.0 implementation
 *  Copyright (C) 1997 Kay Roemer & Arno Puder
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#ifndef __mico_orb_h__
#define __mico_orb_h__

enum InvokeStatus {
    InvokeOk,
    InvokeForward,
    InvokeSysEx,
    InvokeUsrEx
};

enum LocateStatus {
    LocateUnknown,
    LocateHere,
    LocateForward
};

enum RequestType {
    RequestInvoke,
    RequestBind,
    RequestLocate
};

class ObjectAdapter;
class ORBRequest;
class InvokeRec;
struct ORBCallback;


class ORBRequest : public ServerlessObject {
public:
    virtual ~ORBRequest ();

    virtual const char *op_name() = 0;

    virtual void get_in_args (NVList_ptr iparams, Context_ptr &ctx) = 0;
    virtual void get_in_args (DataEncoder *) = 0;

    virtual void get_out_args (Any *res, NVList_ptr oparams,
			       Exception *&ex) = 0;
    virtual void get_out_args (DataEncoder *, Boolean &is_except) = 0;

    virtual void set_out_args (Any *res, NVList_ptr oparams) = 0;
    virtual void set_out_args (Exception *ex) = 0;

    virtual void copy_out_args (ORBRequest *) = 0;
    virtual void copy_in_args (ORBRequest *) = 0;

    virtual const char *type () = 0;

    static ORBRequest* _nil ();
    static ORBRequest* _duplicate (ORBRequest *);
};

MICO_EXPORT_DECL Boolean is_nil (ORBRequest *);
MICO_EXPORT_DECL void release (ORBRequest *);


class ORB : public ServerlessObject {
public:
    typedef CORBA::ULong MsgId;
private:
    typedef vector<ObjectAdapter *> OAVec;
    typedef map<MsgId, InvokeRec *, less<MsgId> > InvokeMap;

    list<string> isa_cache;
    vector<string> _bindaddrs;
    OAVec adapters;
    InvokeMap invokes;
    Dispatcher *disp;
    IOR *tmpl;
    MsgId theid;
    Object_ptr imr, ir, naming;

    ObjectAdapter *get_oa (Object_ptr);
    MsgId new_msgid ();
    void add_invoke (InvokeRec *);
    InvokeRec *get_invoke (MsgId);
    void del_invoke (MsgId);

    Boolean builtin_invoke (MsgId, Object_ptr,
                            ORBRequest *, Principal_ptr);
public:
    typedef SequenceTmpl<Request_var> RequestSeq;
    typedef TSeqVar<RequestSeq> RequestSeq_var;

    typedef SequenceTmpl<OctetWrapper> ObjectTag;
    typedef TSeqVar<ObjectTag> ObjectTag_var;

    ORB (int &argc, char **argv);
    virtual ~ORB ();

    char *object_to_string (Object_ptr);
    Object_ptr string_to_object (const char *);
    Object_ptr ior_to_object (IOR *);
    Boolean is_impl (Object_ptr);

    static char *tag_to_string (const ObjectTag &);
    static ObjectTag *string_to_tag (const char *);

    Status create_list (Long, NVList_ptr &);
    Status create_operation_list (OperationDef *,
    				  NVList_ptr &);
    Status create_named_value (NamedValue_ptr &);
    Status create_exception_list (ExceptionList_ptr &);
    Status create_context_list (ContextList_ptr &);
    Status create_environment (Environment_ptr &);
    
    Status get_default_context (Context_ptr &);

    Status send_multiple_requests_oneway (const RequestSeq &);
    Status send_multiple_requests_deferred (const RequestSeq &);
    Boolean poll_next_response ();
    Status get_next_response (Request_ptr &);

    typedef char *OAid;
    BOA_ptr BOA_init (int &argc, char **argv, const char *boa_id);
    BOA_ptr BOA_instance (const char *id, Boolean create=TRUE);

    typedef char *ObjectId;
    typedef SequenceTmpl<String_var> ObjectIdList;
    typedef ObjectIdList *ObjectIdList_ptr;
    typedef TVar<ObjectIdList> ObjectIdList_var;

    /*
     * XXX which methods raise this exception ?
     * class InvalidName  { };
     */
    ObjectIdList_ptr list_initial_services ();
    Object_ptr resolve_initial_references (const char *id);
    Boolean set_initial_reference (const char *id, Object_ptr obj);

    // begin-not-in-appendix
    virtual TypeCode_ptr create_struct_tc (const char *rep_id,
					   const char *name,
					   const StructMemberSeq &);
    virtual TypeCode_ptr create_union_tc (const char *rep_id,
					  const char *name,
					  TypeCode_ptr disc_type,
					  const UnionMemberSeq &);
    virtual TypeCode_ptr create_enum_tc (const char *rep_id,
					 const char *name,
					 const EnumMemberSeq &);
    virtual TypeCode_ptr create_exception_tc (const char *rep_id,
					      const char *name,
					      const StructMemberSeq &);
    virtual TypeCode_ptr create_alias_tc (const char *rep_id,
					  const char *name,
					  TypeCode_ptr orig_type);
    virtual TypeCode_ptr create_interface_tc (const char *rep_id,
					      const char *name);
    virtual TypeCode_ptr create_string_tc (ULong bound);
    virtual TypeCode_ptr create_sequence_tc (ULong bound,
					     TypeCode_ptr el_type);
    virtual TypeCode_ptr create_recursive_sequence_tc (ULong bound,
						       ULong offset);
    virtual TypeCode_ptr create_array_tc (ULong length,
					  TypeCode_ptr el_type);
    // end-not-in-appendix

    // begin-mico-extension
    const vector<string> &bindaddrs () const;
    void bindaddrs (const vector<string> &);

    Dispatcher *dispatcher ();
    void dispatcher (Dispatcher *);
    IOR *ior_template ();

    void register_oa (ObjectAdapter *);
    void unregister_oa (ObjectAdapter *);

    MsgId next_msgid ();
    
    MsgId invoke_async (Object_ptr, ORBRequest *, Principal_ptr,
			Boolean rply = TRUE, ORBCallback * = 0);
    MsgId locate_async (Object_ptr, ORBCallback * = 0);
    MsgId bind_async (const char *repoid, const ObjectTag &,
		      Address *, ORBCallback * = 0);

    InvokeStatus invoke (Object_ptr &, ORBRequest *, Principal_ptr,
			 Boolean rply = TRUE);
    LocateStatus locate (Object_ptr &);
    LocateStatus bind (const char *repoid, const ObjectTag &,
		       Address *, Object_ptr &);
    Object_ptr bind (const char *repoid, const char *addr = 0);
    Object_ptr bind (const char *repoid, const ObjectTag &,
		     const char *addr = 0);

    void    cancel (MsgId);
    Boolean wait (MsgId, Long tmout = -1);
    Long    wait (vector<MsgId> &, Long tmout = -1);

    void answer_invoke (MsgId, InvokeStatus, Object_ptr, ORBRequest *);
    void answer_bind (MsgId, LocateStatus, Object_ptr);
    void answer_locate (MsgId, LocateStatus, Object_ptr);

    InvokeStatus get_invoke_reply (MsgId, Object_ptr &, ORBRequest *&);
    LocateStatus get_locate_reply (MsgId, Object_ptr &);
    LocateStatus get_bind_reply (MsgId, Object_ptr &);

    RequestType request_type (MsgId);

    ImplementationDef    *get_impl (Object_ptr);
    InterfaceDef         *get_iface (Object_ptr);
    Boolean               is_a (Object_ptr, const char *repoid);
    Boolean               non_existent (Object_ptr);
    // end-mico-extension

    static ORB_ptr _duplicate (ORB_ptr);
    static ORB_ptr _nil ();
};

// XXX we have to implement a real wrapper for class ORB
typedef ObjVar<ORB> ORB_var;

MICO_EXPORT_DECL void release (ORB_ptr);
MICO_EXPORT_DECL Boolean is_nil (ORB_ptr);


// ORB initialization
typedef char *ORBid;
MICO_EXPORT_DECL ORB_ptr ORB_init (int &argc, char **argv, const char *id);
MICO_EXPORT_DECL ORB_ptr ORB_instance (const char *id, Boolean create=TRUE);



class InvokeRec {
public:
    typedef ULong MsgId;
private:
    MsgId myid;
    RequestType type;
    Boolean have_result;
    Object_ptr obj;
    InvokeStatus invoke_stat;
    LocateStatus locate_stat;
    ObjectAdapter *adapter;
    ORBRequest *req;
    ORBCallback *cb;
    ORB_ptr orb;
    ORB::ObjectTag objtag;
public:
    // invoke
    InvokeRec (ORB_ptr, MsgId, ORBRequest *,
	       ORBCallback *, ObjectAdapter * = 0);
    // locate
    InvokeRec (ORB_ptr, MsgId, Object_ptr,
	       ORBCallback *, ObjectAdapter * = 0);
    // bind
    InvokeRec (ORB_ptr, MsgId, const char *,
	       const ORB::ObjectTag &tag, ORBCallback *,
	       ObjectAdapter * = 0);
    virtual ~InvokeRec ();

    MsgId id () const;
    Boolean completed () const;
    ObjectAdapter *oa ();
    void oa (ObjectAdapter *);
    RequestType request_type ();
    ORBRequest *request ();
    const ORB::ObjectTag &tag () const;

    void set_answer_invoke (InvokeStatus, Object_ptr, ORBRequest *);
    void set_answer_bind (LocateStatus, Object_ptr);
    void set_answer_locate (LocateStatus, Object_ptr);

    Boolean get_answer_invoke (InvokeStatus &, Object_ptr &, ORBRequest *&);
    Boolean get_answer_bind (LocateStatus &, Object_ptr &);
    Boolean get_answer_locate (LocateStatus &, Object_ptr &);
};


class ObjectAdapter {
public:
    typedef CORBA::ULong MsgId;

    virtual const char *get_oaid () const = 0;
    virtual Boolean has_object (Object_ptr) = 0;
    virtual Boolean is_local () const = 0;

    virtual Boolean invoke (MsgId, Object_ptr, ORBRequest *,
			    Principal_ptr, Boolean response_exp = TRUE) = 0;
    virtual Boolean bind (MsgId, const char *repoid,
			  const ORB::ObjectTag &,
			  Address *addr) = 0;
    virtual Boolean locate (MsgId, Object_ptr) = 0;
    virtual Object_ptr skeleton (Object_ptr) = 0;
    virtual void cancel (MsgId) = 0;

    virtual ~ObjectAdapter ();
};

#endif // __mico_orb_h__
