// -*- c++ -*-
/*
 *  MICO --- a CORBA 2.0 implementation
 *  Copyright (C) 1997 Kay Roemer & Arno Puder
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#ifndef __mico_transport_impl_h__
#define __mico_transport_impl_h__


class SelectDispatcher : public CORBA::Dispatcher {
    typedef fd_set FDSet;
    typedef struct timeval TimeVal;

    struct FileEvent {
	Event event;
	CORBA::Long fd;
	CORBA::DispatcherCallback *cb;
	CORBA::Boolean deleted;

	FileEvent () {}
	FileEvent (Event _ev, CORBA::Long _fd, CORBA::DispatcherCallback *_cb)
	    : event(_ev), fd(_fd), cb(_cb), deleted(FALSE)
	{}
    };
    struct TimerEvent {
	Event event;
	CORBA::Long delta;
	CORBA::DispatcherCallback *cb;

	TimerEvent () {}
	TimerEvent (Event _ev, CORBA::Long _delta,
		    CORBA::DispatcherCallback *_cb)
	    : event(_ev), delta(_delta), cb(_cb)
	{}
    };
    list<FileEvent> fevents;
    list<TimerEvent> tevents;

    CORBA::Long last_update;
    CORBA::Boolean init;
    CORBA::Long locked;
    CORBA::Boolean modified;
    FDSet curr_wset, curr_rset, curr_xset;

    void lock ();
    void unlock ();
    CORBA::Boolean islocked () const;

    CORBA::Long gettime () const;
    void update_tevents ();
    void handle_tevents ();
    void handle_fevents (FDSet &rset, FDSet &wset, FDSet &xset);
    void update_fevents ();
    void sleeptime (TimeVal &) const;
public:
    SelectDispatcher ();
    virtual ~SelectDispatcher ();
    virtual void rd_event (CORBA::DispatcherCallback *, CORBA::Long fd);
    virtual void wr_event (CORBA::DispatcherCallback *, CORBA::Long fd);
    virtual void ex_event (CORBA::DispatcherCallback *, CORBA::Long fd);
    virtual void tm_event (CORBA::DispatcherCallback *, CORBA::ULong tmout);
    virtual void remove (CORBA::DispatcherCallback *, Event);
    virtual void move (CORBA::Dispatcher *);
    virtual void run (CORBA::Boolean infinite = TRUE);
    virtual CORBA::Boolean idle () const;
};


class TCPTransport : public CORBA::Transport,
		     public CORBA::DispatcherCallback {
    CORBA::Dispatcher *rdisp, *wdisp;
    CORBA::TransportCallback *rcb, *wcb;
    CORBA::Long fd;
    CORBA::Boolean ateof;
    string err;
    InetAddress local_addr, peer_addr;
public:
    TCPTransport (CORBA::Long fd = -1);
    virtual ~TCPTransport ();

    virtual void rselect (CORBA::Dispatcher *, CORBA::TransportCallback *);
    virtual void wselect (CORBA::Dispatcher *, CORBA::TransportCallback *);
    virtual void callback (CORBA::Dispatcher *, CORBA::Dispatcher::Event);

    virtual CORBA::Boolean bind (const CORBA::Address *);
    virtual CORBA::Boolean connect (const CORBA::Address *);
    virtual void close ();
    virtual void block (CORBA::Boolean doblock = TRUE);

    virtual CORBA::Long read (void *, CORBA::Long len);
    virtual CORBA::Long write (void *, CORBA::Long len);

    virtual const InetAddress *addr ();
    virtual const InetAddress *peer ();

    virtual CORBA::Boolean eof () const;
    virtual CORBA::Boolean bad () const;
    virtual string errormsg () const;
};


class TCPTransportServer : public CORBA::TransportServer,
			   public CORBA::DispatcherCallback {
    CORBA::Dispatcher *adisp;
    CORBA::TransportServerCallback *acb;
    CORBA::Long fd;
    string err;
    InetAddress local_addr;
    CORBA::Boolean listening;
    void listen ();
public:
    TCPTransportServer ();
    virtual ~TCPTransportServer ();

    virtual void aselect (CORBA::Dispatcher *,
			  CORBA::TransportServerCallback *);
    virtual void callback (CORBA::Dispatcher *, CORBA::Dispatcher::Event);

    virtual CORBA::Boolean bind (const CORBA::Address *);
    virtual void close ();
    virtual void block (CORBA::Boolean doblock = TRUE);

    virtual TCPTransport *accept ();
    virtual const InetAddress *addr ();

    virtual CORBA::Boolean bad () const;
    virtual string errormsg () const;
};

#ifdef HAVE_SYS_UN_H
class UnixTransport : public CORBA::Transport,
		      public CORBA::DispatcherCallback {
    CORBA::Dispatcher *rdisp, *wdisp;
    CORBA::TransportCallback *rcb, *wcb;
    CORBA::Long fd;
    CORBA::Boolean ateof;
    string err;
    UnixAddress local_addr, peer_addr;
public:
    UnixTransport (CORBA::Long fd = -1);
    virtual ~UnixTransport ();

    virtual void rselect (CORBA::Dispatcher *, CORBA::TransportCallback *);
    virtual void wselect (CORBA::Dispatcher *, CORBA::TransportCallback *);
    virtual void callback (CORBA::Dispatcher *, CORBA::Dispatcher::Event);

    virtual CORBA::Boolean bind (const CORBA::Address *);
    virtual CORBA::Boolean connect (const CORBA::Address *);
    virtual void close ();
    virtual void block (CORBA::Boolean doblock = TRUE);

    virtual CORBA::Long read (void *, CORBA::Long len);
    virtual CORBA::Long write (void *, CORBA::Long len);

    virtual const UnixAddress *addr ();
    virtual const UnixAddress *peer ();

    virtual CORBA::Boolean eof () const;
    virtual CORBA::Boolean bad () const;
    virtual string errormsg () const;
};


class UnixTransportServer : public CORBA::TransportServer,
			    public CORBA::DispatcherCallback {
    CORBA::Dispatcher *adisp;
    CORBA::TransportServerCallback *acb;
    CORBA::Long fd;
    string err;
    UnixAddress local_addr;
    CORBA::Boolean listening;
    void listen ();
public:
    UnixTransportServer ();
    virtual ~UnixTransportServer ();

    virtual void aselect (CORBA::Dispatcher *,
			  CORBA::TransportServerCallback *);
    virtual void callback (CORBA::Dispatcher *, CORBA::Dispatcher::Event);

    virtual CORBA::Boolean bind (const CORBA::Address *);
    virtual void close ();
    virtual void block (CORBA::Boolean doblock = TRUE);

    virtual UnixTransport *accept ();
    virtual const UnixAddress *addr ();

    virtual CORBA::Boolean bad () const;
    virtual string errormsg () const;
};
#endif // HAVE_SYS_UN_H

#endif // __mico_transport_impl_h__
