// -*- c++ -*-
/*
 *  MICO --- a CORBA 2.0 implementation
 *  Copyright (C) 1997 Kay Roemer & Arno Puder
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#ifndef __mico_typecode_h__
#define __mico_typecode_h__

/*
 * missing:
 *  + InterfaceDef and co.
 *  + OperationDef and co.
 *  - TypeCode::Bounds
 *  - TypeCode::BadKind
 *  - UnionMember(Seq)
 *  - StructMember(Seq)
 *  - EnumMemberSeq
 */

// type code
enum TCKind {
    tk_null, tk_void,
    tk_short, tk_long, tk_ushort, tk_ulong,
    tk_float, tk_double, tk_boolean, tk_char,
    tk_octet, tk_any, tk_TypeCode, tk_Principal, tk_objref,
    tk_struct, tk_union, tk_enum, tk_string,
    tk_sequence, tk_array, tk_alias, tk_except,
    tk_recursive = 0xffffffff,
};

class TypeCode : public ServerlessObject {
    TCKind tckind;
    string repoid;
    string tcname;
    vector<string> namevec;
    vector<TypeCode_ptr> tcvec;
    vector<Any *> labelvec;
    TypeCode_ptr content;
    TypeCode_ptr discriminator;
    TypeCode_ptr parent;
    Long defidx;
    Long len;
    Long recurse_depth;
    Boolean isconst;

    void init ();
    void copy (const TypeCode &);
    void free ();

    typedef pair<TypeCode_ptr, TypeCode_ptr> PairTC;
    typedef set<PairTC, less<PairTC> > SetTC;
    typedef map<ULong, pair<ULong, TypeCode_ptr>, less<ULong> > MapPosTC;
    typedef vector<ULong> VecPos;
public:
    // begin-mico-extension
    TypeCode (const string &);
    TypeCode (); // for vector<TypeCode>
    ~TypeCode ();
    TypeCode (TCKind);
    TypeCode (const TypeCode &);
    TypeCode (DataDecoder &);
    TypeCode &operator= (const TypeCode &);
    Boolean from_string (const char *);

    // struct or exception
    TypeCode (const char *rep_id, const char *name,
	      const StructMemberSeq &members, Boolean is_except);
    // union
    TypeCode (const char *rep_id, const char *name,
	      TypeCode_ptr disc_type,
	      const UnionMemberSeq &members);
    // enum
    TypeCode (const char *rep_id, const char *name,
	      const EnumMemberSeq &members);
    // alias
    TypeCode (const char *rep_id, const char *name,
	      TypeCode_ptr orig_type);
    // interface
    TypeCode (const char *rep_id, const char *name);
    // string
    TypeCode (ULong bound);
    // sequence or array
    TypeCode (ULong bound, TypeCode_ptr el_type, Boolean is_array);
    // recursive sequences
    TypeCode (ULong bound, ULong offset);
    // end-mico-extension

    /*
     * XXX which methods raise those exceptions ?
     */
    class BadKind {};
    class Bounds {};

    TCKind kind () const;
    Boolean equal (TypeCode_ptr) const;
    Boolean equaltype (TypeCode_ptr, SetTC * = 0);
    Boolean subtype (TypeCode_ptr, SetTC * = 0);
    TypeCode_ptr max (TypeCode_ptr);

    const char *id () const;
    const char *name () const;

    Long member_index (const Any &);
    Long member_index (const char *);
    ULong member_count () const;
    const char *member_name (ULong idx) const;
    TypeCode_ptr member_type (ULong idx); // const;
    Any *member_label (ULong idx); // const;
    TypeCode_ptr discriminator_type (); // const;
    Long default_index () const;

    ULong length () const;
    
    TypeCode_ptr content_type (); // const;

    Long param_count () const;
    Any *parameter (Long); // const;

    // begin-mico-extension
    Boolean is_recursive_seq ();
    ULong get_recurse_depth ();
    ULong array_size ();
    TypeCode_ptr array_type ();
    Boolean is_constant () const;
    TypeCode_ptr mk_constant ();
    TypeCode_ptr unalias ();

    void encode (DataEncoder &, VecPos * = NULL) const;
    Boolean decode (DataDecoder &, MapPosTC * = NULL, ULong level = 0);

    string stringify () const;
    // end-mico-extension

    static TypeCode_ptr _duplicate (TypeCode_ptr);
    static TypeCode_ptr _nil ();
};
MICO_EXPORT_DECL void release (TypeCode_ptr);
MICO_EXPORT_DECL Boolean is_nil (TypeCode_ptr);

//typedef ObjVar<TypeCode> TypeCode_var;


// begin-mico-extension
class LevelRecord {
public:
    enum Level {
	None, Array, Sequence, Struct, Except, Union,
    };
private:
    TypeCode_ptr _tc;
    Level _level;
public:
    Long i, n, x;

    LevelRecord (Level, TypeCode_ptr, Long n = 0, Long i = 0, Long x = 0);
    LevelRecord (const LevelRecord &);
    LevelRecord ();
    ~LevelRecord ();
    LevelRecord &operator= (const LevelRecord &);

    Level level () const;
    TypeCode_ptr tc ();
};

class TypeCodeChecker {
    TypeCode_ptr _top;
    TypeCode_ptr _tc;
    Boolean done;
    vector<LevelRecord> levelvec;

    LevelRecord *level ();
    Boolean advance ();
    Boolean nexttc ();
    Boolean leave (LevelRecord::Level);
    void enter (const LevelRecord &);
    void tc (TypeCode_ptr);
public:
    TypeCodeChecker ();
    TypeCodeChecker (TypeCode_ptr);
    TypeCodeChecker (const TypeCodeChecker &);
    TypeCodeChecker &operator= (const TypeCodeChecker &);
    ~TypeCodeChecker ();

    TypeCode_ptr tc ();
    Boolean inside (TCKind);

    Boolean basic (TypeCode_ptr);
    Boolean enumeration (Long val);

    Boolean union_begin ();
    Boolean union_selection (Long idx = -1L);
    Boolean union_end ();

    Boolean struct_begin ();
    Boolean struct_end ();

    Boolean except_begin ();
    Boolean except_end ();

    Boolean seq_begin (ULong len);
    Boolean seq_end ();

    Boolean arr_begin ();
    Boolean arr_end ();

    Boolean completed () const;
    void restart (TypeCode_ptr);
    void restart ();
    ULong level_count () const;
};
// end-mico-extension


// XXX add missing _tc_* constants
// XXX add right repository id for _tc_Object

MICO_EXPORT_DECL TypeCode_ptr _tc_null;
MICO_EXPORT_DECL TypeCode_ptr _tc_void;
MICO_EXPORT_DECL TypeCode_ptr _tc_short;
MICO_EXPORT_DECL TypeCode_ptr _tc_long;
MICO_EXPORT_DECL TypeCode_ptr _tc_ushort;
MICO_EXPORT_DECL TypeCode_ptr _tc_ulong;
MICO_EXPORT_DECL TypeCode_ptr _tc_float;
MICO_EXPORT_DECL TypeCode_ptr _tc_double;
MICO_EXPORT_DECL TypeCode_ptr _tc_boolean;
MICO_EXPORT_DECL TypeCode_ptr _tc_char;
MICO_EXPORT_DECL TypeCode_ptr _tc_octet;
MICO_EXPORT_DECL TypeCode_ptr _tc_any;
MICO_EXPORT_DECL TypeCode_ptr _tc_TypeCode;
MICO_EXPORT_DECL TypeCode_ptr _tc_Principal;
MICO_EXPORT_DECL TypeCode_ptr _tc_Context; // extension
MICO_EXPORT_DECL TypeCode_ptr _tc_Object;
MICO_EXPORT_DECL TypeCode_ptr _tc_string;


#endif // __mico_typecode_h__
