// -*- c++ -*-
/*
 *  MICO --- a CORBA 2.0 implementation
 *  Copyright (C) 1997 Kay Roemer & Arno Puder
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#ifndef __mico_var_h__
#define __mico_var_h__

template<class T>
class ObjWrapper {
public:
    static T* nil ();
    static T* duplicate (T *);
    static void release (T *);
};

/*
 * Template to generate _var types for object references [16.3.6]
 */
template<class T>
class ObjVar
{
private:
  T* _ptr;
  T* _ptr_backup;

  void check()
  {
      if (_ptr != _ptr_backup) {
	  // release() works on nil
	  ObjWrapper<T>::release (_ptr_backup);
	  _ptr_backup = _ptr;
      }
  }
  void free()
  {
      check ();
      ObjWrapper<T>::release( _ptr );
      _ptr_backup = _ptr = ObjWrapper<T>::nil();
  }
  void reset( T* ptr )
  {
      free();
      _ptr_backup = _ptr = ptr;
  }
public:
  ObjVar()
  {
      _ptr_backup = _ptr = ObjWrapper<T>::nil();
  }
  ObjVar( T* ptr )
  {
      _ptr = _ptr_backup = ptr;
  }
  ObjVar( const ObjVar<T>& var )
  {
      _ptr_backup = _ptr = ObjWrapper<T>::duplicate (var._ptr);
  }
  ~ObjVar()
  {
      if (_ptr != _ptr_backup)
	  ObjWrapper<T>::release (_ptr_backup);
      ObjWrapper<T>::release (_ptr);
  }
  ObjVar<T>& operator=( const ObjVar<T> &var )
  {
      if (this != &var) {
	  free();
	  _ptr_backup = _ptr = ObjWrapper<T>::duplicate( var._ptr );
      }
      return *this;
  }
  ObjVar<T>& operator=( T* ptr )
  {
      if (_ptr != ptr)
	  reset( ptr );
      return *this;
  }
// g++ const overload problem
#if 0
  operator const T*() const
  {
      return _ptr;
  }
  operator T*&()
  {
      check();
      return _ptr;
  }
#else
  operator T*&() const
  {
      ((ObjVar<T>*)this)->check();
      return (T*&)_ptr;
  }
#endif
  T* operator->() const
  {
      return _ptr;
  }

  T* _retn()
  {
      check ();
      T *_p = _ptr;
      _ptr_backup = _ptr = ObjWrapper<T>::nil();
      return _p;
  }
};


/*
 * Template to generate _var types for data types [16.8.1]
 * See comments for CORBA::String_var concerning _ptr_backup
 */
template<class T>
class TVar
{
private:
  T* _ptr;
  T* _ptr_backup;
  
  void check()
  {
    if( _ptr != _ptr_backup ) {
      if( _ptr_backup )
	delete _ptr_backup;
      _ptr_backup = _ptr;
    }
  }
  void free()
  {
    check();
    if( _ptr )
      delete _ptr;
    _ptr = _ptr_backup = NULL;
  }
  void reset( T* ptr )
  {
    check();
    free();
    _ptr = _ptr_backup = ptr;
  }
public:
  TVar()
  {
    _ptr = _ptr_backup = NULL;
  }
  TVar( T* ptr )
  {
    _ptr = _ptr_backup = ptr;
  }
  TVar( const TVar<T>& var )
  {
    if (!var._ptr)
      _ptr = _ptr_backup = NULL;
    else
      _ptr = _ptr_backup = new T (*var._ptr);
  }
  ~TVar()
  {
    if (_ptr != _ptr_backup)
      delete _ptr_backup;
    if (_ptr)
      delete _ptr;
  }
  TVar<T>& operator=( T* ptr )
  {
    if (_ptr != ptr)
      reset( ptr );
    return *this;
  }
  TVar<T>& operator=( const TVar<T>& var )
  {
    if (this != &var) {
      free();
      if (!var._ptr)
	_ptr = _ptr_backup = NULL;
      else
	_ptr = _ptr_backup = new T (*var._ptr);
    }
    return *this;
  }
// g++ const overload problem
#if 0
  operator const T*() const
  {
    return _ptr;
  }
  operator T*&()
  {
    check();
    return _ptr;
  }
#else
  operator T*&() const
  {
    ((TVar<T>*)this)->check();
    return (T*&)_ptr;
  }
#endif
// g++ const overload problem
#if 0
  operator const T&() const
  {
    assert (_ptr);
    return *_ptr;
  }
  operator T&()
  {
    check();
    assert (_ptr);
    return *_ptr;
  }
#else
  operator T&() const
  {
    ((TVar<T>*)this)->check();
    assert (_ptr);
    return *_ptr;
  }
#endif
  T* operator->() const
  {
    assert (_ptr);
    return _ptr;
  }

  T* _retn()
  {
    check();
    T *_p = _ptr;
    _ptr_backup = _ptr = NULL;
    return _p;
  }
};


/*
 * Template to generate _var types for sequences. The difference
 * to the TVar template is, that operator[] is overloaded as well.
 */
template<class T>
class TSeqVar
{
private:
  T* _ptr;
  T* _ptr_backup;
  
  void check()
  {
    if( _ptr != _ptr_backup ) {
      if( _ptr_backup )
	delete _ptr_backup;
      _ptr_backup = _ptr;
    }
  }
  void free()
  {
    check();
    if( _ptr )
      delete _ptr;
    _ptr = _ptr_backup = NULL;
  }
  void reset( T* ptr )
  {
    free();
    _ptr = _ptr_backup = ptr;
  }
public:
  TSeqVar()
  {
    _ptr = _ptr_backup = NULL;
  }
  TSeqVar( T* ptr )
  {
    _ptr = _ptr_backup = ptr;
  }
  TSeqVar( const TSeqVar<T>& var )
  {
    if (!var._ptr)
      _ptr = _ptr_backup = NULL;
    else
      _ptr = _ptr_backup = new T (*var._ptr);
  }
  ~TSeqVar()
  {
    free();
  }
  TSeqVar<T>& operator=( T* ptr )
  {
    if (_ptr != ptr)
      reset( ptr );
    return *this;
  }
  TSeqVar<T>& operator=( const TSeqVar<T>& var )
  {
    if (this != &var) {
      free();
      if (!var._ptr)
        _ptr = _ptr_backup = NULL;
      else
        _ptr = _ptr_backup = new T (*var._ptr);
    }
    return *this;
  }
// g++ const overload problem
#if 0
  operator const T*() const
  {
    return _ptr;
  }
  operator T*&()
  {
    check();
    return _ptr;
  }
#else
  operator T*&() const
  {
    ((TSeqVar<T>*)this)->check();
    return (T*&)_ptr;
  }
#endif
// g++ const overload problem
#if 0
  operator const T&() const
  {
    assert (_ptr);
    return *_ptr;
  }
  operator T&()
  {
    check();
    assert (_ptr);
    return *_ptr;
  }
#else
  operator T&() const
  {
    ((TSeqVar<T>*)this)->check();
    assert (_ptr);
    return *_ptr;
  }
#endif
  T* operator->() const
  {
    assert( _ptr != NULL );
    return _ptr;
  }
  typename T::ElementType &operator[] (MICO_ULong idx)
  {
    assert( _ptr != NULL );
    return (*_ptr)[ idx ];
  }

  T* _retn()
  {
    check();
    T *_p = _ptr;
    _ptr_backup = _ptr = NULL;
    return _p;
  }
};


/*
 * Template to generate _var types for exceptions. Difference to
 * TVar is that _clone() is used instead of the copy constructor.
 */
template<class T>
class ExceptVar
{
private:
  T* _ptr;
  T* _ptr_backup;
  
  void check()
  {
    if( _ptr != _ptr_backup ) {
      if( _ptr_backup )
	delete _ptr_backup;
      _ptr_backup = _ptr;
    }
  }
  void free()
  {
    check();
    if( _ptr )
      delete _ptr;
    _ptr = _ptr_backup = NULL;
  }
  void reset( T* ptr )
  {
    check();
    free();
    _ptr = _ptr_backup = ptr;
  }
public:
  ExceptVar()
  {
    _ptr = _ptr_backup = NULL;
  }
  ExceptVar( T* ptr )
  {
    _ptr = _ptr_backup = ptr;
  }
  ExceptVar( const ExceptVar<T>& var )
  {
    if (!var._ptr)
      _ptr = _ptr_backup = NULL;
    else
      _ptr = _ptr_backup = var._ptr->_clone();
  }
  ~ExceptVar()
  {
    free();
  }
  ExceptVar<T>& operator=( T* ptr )
  {
    if (_ptr != ptr)
      reset( ptr );
    return *this;
  }
  ExceptVar<T>& operator=( const ExceptVar<T>& var )
  {
    if (this != &var) {
      free();
      if (!var._ptr)
	_ptr = _ptr_backup = NULL;
      else
	_ptr = _ptr_backup = var._ptr->_clone();
    }
    return *this;
  }
// g++ const overload problem
#if 0
  operator const T*() const
  {
    return _ptr;
  }
  operator T*&()
  {
    check();
    return _ptr;
  }
#else
  operator T*&() const
  {
    ((ExceptVar<T>*)this)->check();
    return (T*&)_ptr;
  }
#endif
// g++ const overload problem
#if 0
  operator const T&() const
  {
    assert (_ptr);
    return *_ptr;
  }
  operator T&()
  {
    check();
    assert (_ptr);
    return *_ptr;
  }
#else
  operator T&() const
  {
    ((ExceptVar<T>*)this)->check();
    assert (_ptr);
    return *_ptr;
  }
#endif
  T* operator->() const
  {
    assert (_ptr);
    return _ptr;
  }

  T* _retn()
  {
    check();
    T *_p = _ptr;
    _ptr_backup = _ptr = NULL;
    return _p;
  }
};

#endif
